/* Implementing serial I/O using interrupt driven ring buffers 

	Basically you call putchar to place a char in the buffer.  If the UART is
not sending it will automatically write the char without buffering.  If the
UART is busy sending it will buffer it. When the TX complete interrupt occurs
it will see if more chars are in the buffer.  If there are it will send another
char otherwise it will clear the status flag.  

	You call getchar to recieve a char from the UART.  If there are none left
in the buffer it will wait.  The RX complete interrupt will add new chars to
the buffer.  It is important to note that you must read UDR during a RX 
interrupt otherwise the RXC flag of USR will not be cleared and it will forever
call your interrupt.
*/

#include <avrio.h>
#include <avrreg.h>
#include <avrint.h>

#include "def_avr.h"

/* ring buffer, requires 37 bytes of ram for 16/16 FIFO style buffer */
unsigned char status, in_head, in_tail, out_head, out_tail, in_ring[8],
			  out_ring[8];

#define led_on(v) PORTB &= ~v
#define led_off(v) PORTB |= v

/* outputs a char */
putchar(char c)
{
	unsigned char h;

	if(status & 1) {            /* sending? */
		h = (out_head + 1) & 7;
		while(h == out_tail);
		out_ring[out_head] = c; 				 /* add char */
		out_head = h;
	}else{                      /* not sending we can force it */
		status |= 1;            /* we are sending */
		UDR = c; }
}

putstr(const char *s)
{
	while(*s)
		putchar(*s++);
}

/* Reads one char from the input buffer */
char getchar()
{
	char dummy; 
	
	while(in_tail == in_head); 		/* wait for char */
	dummy = in_ring[in_tail];   	/* get char */
	in_tail = (in_tail + 1) & 7;	/* move tail */
	
	return dummy;
}

/* transmit interrupt */
INTERRUPT(_UARTTX_) tx_int()
{
	/* any more chars? */
	if(out_tail == out_head)
		status &= ~1;           /* no, clear flag */
	else {
		/* send next char */
		UDR = out_ring[out_tail];
		out_tail = (out_tail + 1) & 7; }       /* move tail */
}

/* receieve interrupt */
INTERRUPT(_UARTRX_) rx_int()
{
	/* space in buffer? */
	if(in_head != (in_tail - 1) & 7) {
		in_ring[in_head] = UDR;					/* get char */
		in_head = (in_head + 1) & 7; }			/* move head */
	else
		UDR;                    				/* dummy read */
}

main()
{
	/* setup ring buffer */
	in_tail = in_head = out_head = out_tail = status = 0;

	DDRB = 0xff;
	PORTB = 0xFF;

	UCR     |= RXCIE | TXCIE;     /* enable RX and TX interrupts */
	enable();

	for(;;) {
		putstr("This is a test ......................................... >");
		putchar(getchar());     /* echo stuff using ring buffer */
		putstr("<\r\n"); }
}
