/*
	Implements a serial ring buffer on a 68HC11 board.  In this case
we use the SCI for the communications.  You can use the example api to
access the ring buffer.

	You must change the address of the vector in the INTERRUPT macro to
match your system.  I tested this under MON11 where I changed the vector
manually (CV Command).

Compile with: CC11 ser -pof
*/

#include <6811io.h>
#include <6811reg.h>
#include <6811int.h>

#define SIZE	8				/* size of buffer */
#define	_SCI_	65494			/* SCI interrupt vector (FFD6) */

/* ring buffer */
char in_ring[SIZE], out_ring[SIZE], in_head, in_tail, out_head, out_tail;
char sending;


/* SCI interrupt handler */
INTERRUPT(_SER_) sci_int()
{
	/* transmit interrupt? */
	if(SCSR & 0x40) {
		SCSR;

		/* fetch new char and send it */
		if(out_head != (out_tail + 1) % SIZE) {
			out_tail = (out_tail + 1) % SIZE;
			SCDR = out_ring[out_tail]; } 
		else {
			/* disable tx int */
			SCCR2 &= ~0x40;
			sending = 0; } }

	/* rx interrupt? */
	if(SCSR & 0x20) {
		SCSR;						/* clear interrupt flags */

		/* room in buffer? */
		if(in_head != (in_tail - 1) % SIZE) {
			in_ring[in_head] = SCDR;
			in_head = (in_head + 1) % SIZE; }
		else
			SCDR; }					/* clear register */
}

/* user routines */
putchar(char c)
{
	if(!sending) {				/* stuff first char */
		sending = 1;
		SCDR	= c;	
		SCCR2	|= 0x40; }	/* enable TX */
	else {
		/* buffer the char */
		while (out_head == (out_tail - 1) % SIZE);	/* wait for space */

		out_ring[out_head] = c;
		out_head = (out_head + 1) % SIZE; }
}

char getchar()
{
	char t;

	while(in_head == (in_tail + 1) % SIZE);		/* wait for char */

	/* get char from buffer */
	in_tail = (in_tail + 1) % SIZE;
	t = in_ring[in_tail];

	return t;
}

putstr(char *s)
{
	while (*s)
		putchar(*s++);
}

main()
{
	/* setup ring buffer */
	in_head = out_head = 1;
	in_tail = out_tail = sending = 0;

	/* set intvec under MON11 */
	pokew(0, &sci_int);

	/* setup hardware */
	SCCR2 = 0x6C;				/* setup SCI */
	enable();

	/* loop */
	for(;;) {
		putstr("Char is.........<"); putchar(getchar()), putstr("> \n\r"); }
}
