/* Micro-C Task Swapper

	Everything you need to perform task swapping with message queues for
the tasks, in a C callable interface.

	This program demos how to perform simple task swapping, in relatively
small ammount of code.  Also provided are message queues, which allows the
tasks to intercommunicate.

	To setup the tasks you must add their stack pointer (SP) to a list,
then 'stack' the address of their function on that 'stack'.  See main() for
one method of setting this up.

	Once your tasks are loaded, you call the start() function and away it 
goes.  When programming tasks, all you have to do is call 'swap()' to give 
control to another task.  I suggest putting a 'swap()' at the top of all your 
loops, and when you are idle.  This will allow for the least ammount of 
latency between tasks.

	You can also send messages, however they must be kept short as the 
buffer is quite small.  The code provided will handle packing a header
on the message (the header contains the size, and sender only), and stack
it in the message queue.  The message queue is a FIFO buffer.

	Compile:  ccXXX task -pof
*/

/* 68HC12 includes 
#include <6812int.h>
#include <6812reg.h>
#include <6812io.h>
*/

/* 68HC11 includes 
#include <6811int.h>
#include <6811reg.h>
#include <6811io.h>
*/

/* 8096 includes */
#include <8096io.h>
#include <8096int.h>

/* 8086 includes */
// #include <stdio.h>

#define TASKS 		4				/* number of tasks */
#define MSG_SIZE	32				/* size of message queue (per task) */

/* task info */
unsigned *task_list[TASKS];
unsigned current_task;

/* message handling */
unsigned char messages[TASKS][MSG_SIZE];

/* get message, returns bytes read from message queue (0 = no msg) */
int get_message(int task, char *dst)
{
	int size;

	if(size = messages[task][0]){
		/* there are bytes waiting */
		memcpy(dst, &messages[task][1], size);
		memcpy(&messages[task][0], &messages[task][size], MSG_SIZE - size); }

	return size;
}

/* send a message returns 0 on success */
int send_message(int dest, int from, char *data, int size)
{
	int t;

	/* find end of queue for dest stack */
	for(t = 0; t < (MSG_SIZE - 2) && messages[dest][t] != 0; t += messages[dest][t]);

	/* enough space? */
	if(t + size + 2 > MSG_SIZE)
		return -1;
	else {
		messages[dest][t] = size + 2;
		messages[dest][t + 1] = from;
		memcpy(&messages[dest][t + 2], data, size);
		return 0; }
}

/* pick the swapper module */
// #include "12swap.c"					/* 68HC12 cpu */
// #include "11swap.c"					/* 68HC11 cpu */
// #include "86swap.c"					/* 8086 cpu */
#include "96swap.c"						/* 8096 cpu */


/* test tasks (example only) */
/* this is a define, since all of our tasks follow a model */
#define task(name, x) 												\
void name (void) 													\
{																	\
	char c, buf[1 + MSG_SIZE];										\
																	\
	for(;;){														\
		if(get_message(x, buf)) 									\
			printf("Task #%5u: Message from task %u <%s>\n", x, buf[0], &buf[1]); \
																	\
		if(c = chkch()){ 											\
			sprintf(buf, "Char Sent <%c>", c);						\
			while (send_message((x + 1) % 4, x, buf, 13))			\
				swap();	}											\
		swap(); }													\
}

task(_task_1, 0)
task(_task_2, 1)
task(_task_3, 2)
task(_task_4, 3)


/* stack for the tasks */
char stack[1024]; 

main()
{
	/* setup tasks */
	add_task(&_task_1, 0, stack + 1024);
	add_task(&_task_2, 1, stack + 768);
	add_task(&_task_3, 2, stack + 512);
	add_task(&_task_4, 3, stack + 256);

	current_task = 0;							/* first task */
	memset(messages, 0, sizeof(messages));		/* clear message queue */

	/* start session */
	start();
}
